<?php
// $Id$

/**
 * @file
 * Implements the process of parsing and rolling dice, moved here to save on code footprint when dice results are being viewed, not rolled.
 */

/**
 * Dice processing related functions.
 */

/**
 * Given user commands and notes, split command set into individual commands for processing. Then save and return roll set information.
 */
function _dice_roller_parse_roll($roll_command, $roll_notes, $pid, $origin) {
  $command_set = explode(';', trim($roll_command));
  $results = array();
  $commands = array();
  foreach ($command_set as $command) {
    $roll = _dice_roller_parse_single_roll($command);
    if ($roll !== FALSE) {
      $results[] = $roll;
      $commands[] = $command;
    }
    else if(!empty($command)) {
      drupal_set_message(t("There was an error in rolling your dice selection @command.", array('@command' => $command)));    
    }
  }
  
  if (!empty($results)) {
    foreach ($results as $index => $result) {
      $dids[] = _dice_roller_insert_roll($commands[$index], $results[$index], $roll_notes, $pid, $origin);
    }
    return $dids;
  }
  return FALSE;
}

/**
 * Given a single user roll command, parse command according to available behaviors, and return roll results.
 */
function _dice_roller_parse_single_roll($roll_command) {
  $error = FALSE;
  $command_set = array();
  $number_of_dice = 0;
  $max_sides = 0;
  
  // Prepare roll by removing all whitespace.  
  $roll_command = trim($roll_command);
  $roll_command = str_replace(' ', '', $roll_command);
  $split1 = explode('d', $roll_command); // NO APPROVED FUNCTIONS MAY HAVE A 'd' IN THEM!

  if (count($split1) == 2) {
    if (is_numeric($split1[0])) {
      $number_of_dice = $split1[0];
      if (is_numeric($split1[1])) {
        $max_sides = $split1[1];
      }
      else {
        // This split could be ., +, or -.
        $plus = strpos($split1[1], '+');
        $minus = strpos($split1[1], '-');
        $command_found = strpos($split1[1], '.');
        if ($plus !== FALSE) {
          $split2 = explode('+', $split1[1]);
          if (is_numeric($split2[0])) {
            $max_sides = $split2[0];
            if (is_numeric($split2[1])) {
              // Set the each function.
              $command_set['each'] = array($split2[1]);
            }
            else {
              $error = TRUE; // Modifier to add - not a number.
            }
          }
          else {
            $error = TRUE; // Type of dice to roll - not a number.
          }
        }
        elseif ($minus !== FALSE) {
          $split2 = explode('-', $split1[1]);
          if (is_numeric($split2[0])) {
            $max_sides = $split2[0];
            if (is_numeric($split2[1])) {
              // Set the each function.
              $command_set['each'] = array(-1*($split2[1]));
            }
            else {
              $error = TRUE; // Modifier to add - not a number.
            }
          }
          else {
            $error = TRUE; // Type of dice to roll - not a number.
          }
        }
        elseif ($command_found !== FALSE) {
          $split2 = explode('.', $split1[1]);
          if (is_numeric($split2[0])) {
            $max_sides = $split2[0];
            foreach ($split2 as $key => $this_command) {
              if ($key > 0) { // skip the first value - the Type of die.
                $left_parenthesis_position = strpos($this_command, '(');
                $right_parenthesis_position = strpos($this_command, ')');
                if (!($left_parenthesis_position == 0 || $right_parenthesis_position != (drupal_strlen($this_command) - 1) || $left_parenthesis_position > $right_parenthesis_position || $left_parenthesis_position === FALSE || $right_parenthesis_position === FALSE)) {
                  $command_name = drupal_strtolower(drupal_substr($this_command, 0, $left_parenthesis_position));
                  $command_parameters = explode(',', drupal_substr($this_command, ($left_parenthesis_position + 1), ($right_parenthesis_position - (1 + $left_parenthesis_position))));
                  if (_dice_roller_check_function($command_name, $command_parameters) === TRUE) {
                    //At this point, the name is legit, and the parameters are all numeric.
                    $command_set[$command_name] = $command_parameters;
                    // Array of Arrays, indexed by function name.
                    // Duplicated functions overwrite each other the later they appear in the string.
                  }
                  else {
                    // Ignore this function. Wrong name, or bad parameters.
                  }
                }
                else {
                  // Ignore this function. Parenthesis are all messed up.
                }
              }
            }
          }
          else {
            $error = TRUE; // Type of dice to roll - not a number.
          }
        }
      }
    }
    else {
      $error = TRUE; // # of dice to roll - not a number.
    }
  }
  else {
    $error = TRUE; // Too many 'd's to be legit.
  }

  // Ok - are we still not in err state?
  if ($error !== TRUE && $number_of_dice > 0) {
    return _dice_roller_make_roll($number_of_dice, $max_sides, $command_set);
  }
  else {
    return FALSE;
  }
}

/**
 * Check if the given behavior requested is a legitimate command.
 */
function _dice_roller_check_function($command_name, $command_parameter) {
/*
 * Options are:
 * takeHighest  take highest X rolls
 * 10d10.takeHighest(4)  takes highest 4
 * takeLowest  take lowest x rolls
 * 10d10.takeLowest(4)  take lowest 4
 * minroll  minimum low range
 * 10d10.minroll(4)  no die roll lower than 4
 * extra  roll an extra die when roll greater or equal to X
 * 10d10.extra(9)  roll an extra die when a die roll is 9 or higher.
 * open  same as extra but roll extra dice until a die is not greater or equal to X (even the extra roll).
 * 10d10.open(9)  roll extra dice until a die roll is not 9 or higher.
 * each  apply X value to all dice
 * 10d10.each(2)]  add 2 to every die roll
 * hits(hitmin)  roll the dice and count the hits
 * 5d6.hits(5)  (count the hits 5 or greater)  [5,5,6,2,6] = (4)
 * hitsopen(hitmin, reroll)  roll the dice and count the hits, adding an extra die for any value equal or greater than reroll
 * 7d6.hitsopen(5,6)  (count the hits 5 or greater, extra roll on 6)  [[6, 6, 3],4,1,4,4,3,5] = (3)
 */
  $approved_comands = array(
    'takehighest',
    'takelowest',
    'minroll',
    'extra',
    'open',
    'each',
    'hits',
    'hitsopen',
  );

  if (!in_array($command_name, $approved_comands)) {
    return FALSE;
  }

  foreach ($command_parameter as $this_parameter) {
    if (!is_numeric($this_parameter)) {
      return FALSE;
    }
  }
  return TRUE;
}

/**
 * Execute user selected behaviors, and return result set.
 */
function _dice_roller_make_roll($number_of_dice, $max, $command_set) {
  // PRE ROLL FUNCS:  minroll.
  $min = 1;
  if (isset($command_set['minroll'][0]) && $command_set['minroll'][0] < $max) {
    $min = $command_set['minroll'][0];
  }

  // MID ROLL FUNCS:  each. extra, open, hitsopen.
  $roll_extra_dice_limit = 'NA';
  $roll_open_dice_limit = 'NA';
  $add_to_each = 0;
  if (isset($command_set['each'][0])) {
    $add_to_each = $command_set['each'][0];
  }
  if (isset($command_set['extra'][0])) {
    $roll_extra_dice_limit = $command_set['extra'][0];
  }
  if (isset($command_set['open'][0])) {
    $roll_open_dice_limit = $command_set['open'][0];
  }
  if (isset($command_set['hitsopen'][1])) {
    $roll_open_dice_limit = $command_set['hitsopen'][1];
  }

  $roll_set = array();
  $extras_set = array();
  for ($i = 0; $i < $number_of_dice; $i++) {
    $this_roll = 0;
    $this_roll_extras = array();
    $this_roll = mt_rand($min, $max) + $add_to_each;

    if (is_numeric($roll_extra_dice_limit) && ($this_roll >= $roll_extra_dice_limit)) {
      $this_roll_extras[] = mt_rand($min, $max) + $add_to_each;
    }
    
    $this_extra = 0;
    if (is_numeric($roll_open_dice_limit) && ($this_roll >= $roll_open_dice_limit)) {
      $this_extra = mt_rand($min, $max) + $add_to_each;
      while ($this_extra >= $roll_open_dice_limit) {
        $this_roll_extras[] = $this_extra;
        $this_extra = mt_rand($min, $max) + $add_to_each;
      }
      $this_roll_extras[] = $this_extra;
    }
    $roll_set[] = $this_roll;
    $extras_set[] = $this_roll_extras;
  }

  // POST ROLL FUNCS:  hits, hitsopen.
  $count_hits_dice_limit = 'NA';
  if (isset($command_set['hits'][0])) {
    $count_hits_dice_limit = $command_set['hits'][0];
  }
  if (isset($command_set['hitsopen'][0])) {
    $count_hits_dice_limit = $command_set['hitsopen'][0];
  }

  $roll_summary_hits = '';
  $roll_summary_sums = 0;
  foreach ($roll_set as $index => $roll) {
    if (is_numeric($count_hits_dice_limit)) {
      if ($roll >= $count_hits_dice_limit) {
        $roll_summary_hits++;
      }
      foreach ($extras_set[$index] as $extra) {
        if ($extra >= $count_hits_dice_limit) {
          $roll_summary_hits++;
        }
      }
    }

    $roll_summary_sums += $roll;
    foreach ($extras_set[$index] as $extra) {
      $roll_summary_sums += $extra;
    }
  }

  // POST POST ROLL FUNCS: takeHighest, takeLowest.
  $take_highest_dice_limit = 'NA';
  $take_lowest_dice_limit = 'NA';
  if (isset($command_set['takehighest'][0])) {
    $take_highest_dice_limit = $command_set['takehighest'][0];
  }
  if (isset($command_set['takelowest'][0])) {
    $take_lowest_dice_limit = $command_set['takelowest'][0];
  }

  // If either of these is set, we need to flatten the extra results into the main results so we can sort it.
  $flat_results_set = array();
  $roll_summary_taken = '';
  if (is_numeric($take_highest_dice_limit) || is_numeric($take_lowest_dice_limit)) {
    foreach ($roll_set as $index => $roll) {
      $flat_results_set[]  = $roll;
      foreach ($extras_set[$index] as $extra ) {
        $flat_results_set[] = $extra;
      }
    }
    $take_results_set = array();
    if (is_numeric($take_highest_dice_limit)) {
      rsort($flat_results_set);
      for ($i = 0; $i < $take_highest_dice_limit; $i++) {
        $take_results_set[] = $flat_results_set[$i];
      }
    }
    if (is_numeric($take_lowest_dice_limit)) {
      sort($flat_results_set);
      for ($i = 0; $i < $take_lowest_dice_limit; $i++) {
        $take_results_set[] = $flat_results_set[$i];
      }
    }
    $roll_summary_taken = implode(', ', $take_results_set);
  }

  //Done with processing - Create output string.
  $end_results_string = '';
  foreach ($roll_set as $index => $this_roll) {
    $extra_rolls_string = implode(', ', $extras_set[$index]);
    if (!empty($extra_rolls_string)) {
      $roll_set[$index] = $this_roll . ', [' . $extra_rolls_string . ']';
    }
  }

  $end_results_string = implode(', ', $roll_set );
  if (!empty($roll_summary_taken)) {
    $end_results_string .= ' = ' . $roll_summary_taken;
  }
  elseif (!empty($roll_summary_hits)) {
    $end_results_string .= ' = ' . $roll_summary_hits;
  }
  elseif (!empty($roll_summary_sums)) {
    $end_results_string .= ' = ' . $roll_summary_sums;
  }
  else {
    $end_results_string .= ' = ?';
  }
  return $end_results_string;
}